#include "TI92dasm.h"


List ListLabels;
List ListImported;
List ListEntryPoints;
List ListStopPoints;
List ListDisas;
List ListFirst;
tBuffer * Buffer;
int FindStrings = 0;
int DumpSymbols = 0;

void ComputeFlags()
{
   byte * Flags = Buffer->Flags;
   word i, Size = Buffer->Size;
   tListItem * Current = NULL;
   byte * RenduFlags, * RenduCode;

   while((Current = ListEntryPoints.GetNext(Current)) != NULL)
   {
      RenduCode = Buffer->File + Current->Offset;
      RenduFlags = Buffer->Flags + Current->Offset;
      while(RenduCode = Disasm(RenduCode));
      ListEntryPoints.Rm(NULL);
      Current = NULL;
   }


   char Tmp[100];
   for(i = 0 ; i < Size ; i++)
   {
      if((Flags[i] & TO_OUTPUT) && !(Flags[i] & TO_DISAS))
      {
         int StrSize = 0, j = 0;

         if(Current = ListImported.Search(i))
         {
            if(Current->PointTo & 0x8000)
            {
               sprintf(Tmp, "dc.w\t%s", Current->Label);
               ListDisas.Add(TypeData, i, Tmp, 0);
               i+=1;
            }
            else
            {
               sprintf(Tmp, "dc.l\t%s", Current->Label);
               ListDisas.Add(TypeData, i, Tmp, 0);
               i+=3;
            }
            ListImported.Rm(Current);
            continue;
         }


         if(ListLabels.Search(i) && FindStrings)
         {
            while(!iscntrl(Buffer->File[i + StrSize])
                  && Buffer->File[i + StrSize])
            {
               if(StrSize && ListLabels.Search(i + StrSize)) break;
               StrSize++;
            }
            if(!Buffer->File[i + StrSize] && (StrSize > 4))
            {
               char * addstr = "dc.b\t\"";
               char * str = (char *)Errmalloc(StrSize + 1 + strlen(addstr) + 1 + 2);
               int ideb = i;
               strcpy(str, addstr);
               int slen = strlen(str);
               for(j = 0 ; j < StrSize ; j++)
               {
                  str[j + slen] = Buffer->File[i++];
               }
               str[j++ + slen] = '\"';
               str[j++ + slen] = ',';
               str[j++ + slen] = '0"';
               str[j + slen] = 0;

               ListDisas.Add(TypeData, ideb, str, 0);

               continue;
            }
         }

         strcpy(Tmp, "dc.b\t$");
         sprintf(Tmp + strlen(Tmp), "%02X", Buffer->File[i]);
         ListDisas.Add(TypeData, i, Tmp, 0);
      }
   }
}


void * Errmalloc(size_t size)
{
   static size_t TotalAllocated = 0;
   static int AlreadyDumped = 0;
   void * Ret = malloc(size);
   if(Ret == NULL)
   {
      void DumpSymboles();

      printf("Error allocating memory, already %ld bytes allocated.\n", TotalAllocated);
      printf("And %d Disas.\n", ListDisas.card());
      if(!AlreadyDumped)
      {
         AlreadyDumped = 1;
         printf("Erreur !!!!\n");
         DumpSymboles();
      }
      exit(1);
   }
   TotalAllocated += size;
   return Ret;
}


char * GetLabelName(word pOffset)
{
   static int Num = 0;
   char * Ret;
   static char Tmp[5];
   tListItem * Item = NULL;

   Item = ListLabels.Search(pOffset);
   if(Item != NULL)
   {
      Ret = Item->Label;
   }
   else
   {
      sprintf(Tmp, "L%03X", Num);
      ListLabels.Add(TypeLabel, pOffset, Tmp, 0);
      Ret = Tmp;
      Num++;
   }
   return Ret;
}

void DumpSymboles()
{
 List Output;
 if(DumpSymbols)
 {
      printf("Dump of the program symbol table\n");
      printf(" Labels\n");
      ListLabels.SortFollowingToOffset();
      ListLabels.Print( "   ");
      printf(" Imported symboles\n");
      ListImported.SortFollowingToOffset();
      ListImported.Print("   ");
      printf(" Entry points\n");
      ListEntryPoints.SortFollowingToOffset();
      ListEntryPoints.Print("   ");
      printf(" Disassembled code\n");
      ListDisas.SortFollowingToOffset();
      ListDisas.Print("   ");
      printf("\n\n");
   }
   Output.Merge(ListFirst);
   //ListFirst.Free();
   Output.Merge(ListLabels);
   //ListLabels.Free();
   Output.Merge(ListDisas);
   //ListDisas.Free();
   Output.SortFollowingToOffset();
   Output.Print("");
}

int main(int argc, char ** argv)
{
 word Offset;
 FILE * Infos;

 clrscr();

 if(!(Buffer = GetFile(argv[1])))
 {
    printf("Error on file openning.\n");
    return 1;
 }

 if(Infos = fopen(argv[2],"r"))
 {
    char Type[100];
    word Num;
    while(!feof(Infos))
    {
       int PrisEnCompte = 0;

       fscanf(Infos, "%s =%X\n", Type, &Num);
       if(!strcmp(Type, "EntryPoint"))
       {
          ListEntryPoints.Add(TypeEntryPoint, Num, GetLabelName(Num), 0);
          PrisEnCompte = 1;
       }
       else if(!strcmp(Type, "StopPoint"))
       {
          ListStopPoints.Add(TypeEntryPoint, Num, GetLabelName(Num), 0);
          PrisEnCompte = 1;
       }
       else if(!strcmp(Type, "FindStrings"))
       {
          FindStrings = 1;
          PrisEnCompte = 1;
       }
       else if(!strcmp(Type, "DumpSymbols"))
       {
          DumpSymbols = 1;
          PrisEnCompte = 1;
       }

       if(PrisEnCompte) fprintf(stderr, "Got from .nfo : %s at $%04X\n", Type, Num);
    }
    fclose(Infos);
 }

 if(!(!strncmp("68kP", Buffer->File + 4, 4) || !strncmp("68kL", Buffer->File + 4, 4)))
 {
    printf("Error : not a DoorsOS file.\n");
    return 1;
 }

 if(!strncmp("68kL", Buffer->File + 4, 4))
 {
    ListFirst.Add(TypeTexte, 0, "\txdef _library\n_library:", 0);
 }

 Offset = ToWord(Buffer->File + 0xA);
 if(Offset)
 {
    ListLabels.Add(TypeLabel, Offset, "_comment", 0);
    ListFirst.Add(TypeTexte, 0, "\txdef _comment", 0);
 }
 Offset = ToWord(Buffer->File + 0xC);
 if(Offset)
 {
    ListLabels.Add(TypeLabel, Offset, "_main", 0);
    ListEntryPoints.Add(TypeEntryPoint, Offset, "_main", 0);
    ListFirst.Add(TypeTexte, 0, "\txdef _main", 0);
 }
 Offset = ToWord(Buffer->File + 0xE);
 if(Offset)
 {
    ListLabels.Add(TypeLabel, Offset, "_exit", 0);
    ListEntryPoints.Add(TypeEntryPoint, Offset, "_exit", 0);
    ListFirst.Add(TypeTexte, 0, "\txdef _exit", 0);
 }

 if(ToWord(Buffer->File + 0x10) & 1) ListFirst.Add(TypeTexte, 0, "\txdef _ti92plus\n_ti92plus:", 0);
 if(ToWord(Buffer->File + 0x10) & 2) ListFirst.Add(TypeTexte, 0, "\txdef _ti89\n_ti89:", 0);

 ListFirst.Add(TypeTexte, 0, "\n", 0);

 ListImported = ReadImported(Buffer);

 ListFirst.Add(TypeTexte, 0, "\n", 0);
 ListFirst.Add(TypeTexte, 65535, "\tend", 0);

 ComputeFlags();
 DumpSymboles();

 if(ListImported.card() != 0)
    fprintf(stderr, "There might have been at least one error.\n");

 return 0;
}