#include "TI92dasm.h"

#define NbOpcodes 118

#define Dreg 1     // Dn
#define Areg 2     // An
#define Aind 4     // (An)
#define Aipi 8     // (An)+
#define Apdi 16    // -(An)
#define Ad16 32    // x.w(An)
#define Ad8r 64    // x.b(An, Xn.z)
#define absw 128   // x.w
#define absl 256   // x.l
#define PC16 512   // x.w(pc)
#define PC8r 1024  // x.b(pc, Xn.z)
#define immd 2048  // SizeOperandes bytes
#define imm3 4096  // 3 bits
#define imm4 8192  // 4 bits
#define imm8 16384 // 8 bits
#define Sr   32768 // sr
#define RegLst 65536
#define Usp    131072

#define IsEAg 262144   // gauche
#define iseag(x) (x&IsEAg)

#define IsEAd 524288   // droite
#define isead(x) (x&IsEAd)

#define IsRel 1048576  // pour bcc...

#define Dreg2 2097152  // DR
#define Areg2 4194304  // AR


#define StandSrc  (Dreg+Areg+Aind+Aipi+Apdi+Ad16+Ad8r+absw+absl+PC16+PC8r+immd)
#define StandDest (Dreg+Areg+Aind+Aipi+Apdi+Ad16+Ad8r+absw+absl)

#define NoSize 0
#define Byte   1
#define Word   2
#define Long   4
#define GetSize 8
#define AllS   (Byte + Word + Long)

char * CCodes[] = {"ra", "f", "hi", "ls", "cc", "cs", "ne", "eq", "vc", "vs", "pl", "mi", "ge", "lt", "gt", "le"};

typedef struct sInstructionList
{
 word CleanOpcode;
 word Mask;

 word NbOperandes;
 word SizeOperandes;

 long AllowedForSource;
 long AllowedForDest;

 char * Mnemonic;
// char * (*Function)(word pOffset, struct sInstructionList * InstrStruct);
} tInstructionList;

tInstructionList InstructionsTable[] = {
#include "table.h"
};

int Size;
long EAType;
int Dest;
tListItem * ItemToDelete;
word Referenced;

byte * DecodeEA(byte * Code, int Mode, int Reg, char * str)
{
   longword Plus;
   word Plusw;
   tListItem * Item;
   switch(Mode)
   {
      case 0: sprintf(str, "d%d", Reg); EAType = Dreg;
              break;
      case 1: sprintf(str, "a%d", Reg); EAType = Areg;
              break;
      case 2: sprintf(str, "(a%d)", Reg); EAType = Aind;
              break;
      case 3: sprintf(str, "(a%d)+", Reg); EAType = Aipi;
              break;
      case 4: sprintf(str, "-(a%d)", Reg); EAType = Apdi;
              break;
      case 5: sprintf(str, "%d(a%d)", (signed int)ToWord(Code), Reg); EAType = Ad16;
              Code += 2;
              break;
      case 6: Plusw = ToWord(Code); EAType = Ad8r;
              Code += 2;
              if(Plusw & (128*256))
                 sprintf(str, "%d(a%d,a%d.%c)", (signed int)((signed char)(Plusw & 0xFF)), Reg, (Plusw >> 12) & 7, (Plusw & 2048)?'l':'w');
              else
                 sprintf(str, "%d(a%d,d%d.%c)", (signed int)((signed char)(Plusw & 0xFF)), Reg, (Plusw >> 12) & 7, (Plusw & 2048)?'l':'w');
              break;
      case 7: switch(Reg)
              {
                 case 0: if(Item = ListImported.Search(Code - Buffer->File))
                         {
                            if(!(Item->PointTo & 0x8000))
                            {
                               fprintf(stderr, "Erreur : RAM call should be WORD.\n");
                               exit(1);
                            }
                            strcpy(str, Item->Label);
                            ItemToDelete = Item;
                         }
                         else
                         {
                            sprintf(str, "$%X", ToWord(Code));
                         }
                         EAType = absw;
                         Code += 2;
                         break;
                 case 1: if(Item = ListImported.Search(Code - Buffer->File))
                         {
                            Code += 4;
                            strcpy(str, Item->Label);
                            if(Item->Type == TypeReloc)
                               Referenced = Item->PointTo;
                            ItemToDelete = Item;
                         }
                         else
                         {
                            Plus = ToWord(Code); Code += 2;
                            sprintf(str, "$%lX", (Plus << 16) + ToWord(Code));
                            Code += 2;
                         }
                         EAType = absl;
                         break;
                 case 2: Plus = Code - Buffer->File + (signed int)ToWord(Code);
                         sprintf(str, "%s(pc)", GetLabelName(Plus)); EAType = PC16;
                         Referenced = Plus;
                         Code += 2;
                         break;
                 case 3: Plusw = ToWord(Code); EAType = PC8r;
                         Code += 2;
                         if(Plusw & (128*256))
                            sprintf(str, "%d(pc,a%d.%c)", (signed int)((signed char)(Plusw & 0xFF)), (Plusw >> 12) & 7, (Plusw & 2048)?'l':'w');
                         else
                            sprintf(str, "%d(pc,d%d.%c)", (signed int)((signed char)(Plusw & 0xFF)), (Plusw >> 12) & 7, (Plusw & 2048)?'l':'w');
                         break;
                 case 4: switch(Size)
                         {
                            case Byte: sprintf(str, "#$%X", ToWord(Code) & 0xFF); Code += 2; break;
                            case Word: if(Item = ListImported.Search(Code - Buffer->File))
                                       {
                                          if(!(Item->PointTo & 0x8000))
                                          {
                                             fprintf(stderr, "Erreur : RAM call should be WORD.\n");
                                             exit(1);
                                          }
                                          strcpy(str, "#");
                                          strcat(str, Item->Label);
                                          ItemToDelete = Item;
                                       }
                                       else
                                       {
                                          sprintf(str, "#$%X", ToWord(Code));
                                       }
                                       Code += 2;
                                       break;
                            case Long: if(Item = ListImported.Search(Code - Buffer->File))
                                       {
                                          Code += 4;
                                          strcpy(str, "#");
                                          strcat(str, Item->Label);
                                          ItemToDelete = Item;
                                       }
                                       else
                                       {
                                          Plus = ToWord(Code);
                                          Code += 2;
                                          sprintf(str, "#$%lX", (Plus << 16) + ToWord(Code));
                                          Code += 2;
                                       }
                                       break;
                         }
                         EAType = immd;
                         break;
              }
              break;
   }
   return Code;
}

byte * ReadEAd(byte * Code, word pOpcode, char * str)
{
 int Mode = (pOpcode >> 3) & 7;
 int Reg = pOpcode & 7;
 DecodeEA(Code, Mode, Reg, str);
}

byte * ReadEAg(byte * Code, word pOpcode, char * str)
{
 int Mode = (pOpcode >> 6) & 7;
 int Reg = (pOpcode >> 9) & 7;
 return DecodeEA(Code, Mode, Reg, str);
}

byte * Decode(char * str, byte * Code, word Opcode, long Allowed, int * Found)
{
   int isrel = 0;
   if(Allowed & IsRel)
   {
      Allowed &= ~IsRel;
      isrel = 1;
   }

   if(isead(Allowed))
   {
      Code = ReadEAd(Code, Opcode, str);
      if(!(EAType & Allowed)) *Found = 0;
   }
   else if(iseag(Allowed))
   {
      Code = ReadEAg(Code, Opcode, str);
      if(!(EAType & Allowed)) *Found = 0;
   }
   else if(Allowed == Areg)
   {
      sprintf(str, "a%d", (Opcode >> 9) & 7);
   }
   else if(Allowed == Dreg)
   {
      sprintf(str, "d%d", (Opcode >> 9) & 7);
   }
   else if(Allowed == Areg2)
   {
      sprintf(str, "a%d", Opcode & 7);
   }
   else if(Allowed == Dreg2)
   {
      sprintf(str, "d%d", Opcode & 7);
   }
   else if(Allowed == Sr)
   {
      strcpy(str, "sr");
   }
   else if(Allowed == RegLst)
   {
      word Tmp1 = ToWord(Code);
      Code += 2;

      if(!(((Opcode >> 3) & 7) == 4))
      {
         word Tmp2 = 0;
         word pui1, pui2, i;
         for(pui1 = 1, pui2 = 32768, i = 0 ; i <= 15 ; pui1 *= 2, pui2 /= 2, i++)
            if(Tmp1 & pui1) Tmp2 |= pui2;
         Tmp1 = Tmp2;
      }

      int i;
      word Puiss;
      for(Puiss = 32768, i = 0 ; i <= 7 ; i++, Puiss /= 2)
         if(Tmp1 & Puiss) sprintf(str + strlen(str), "d%d/", i);
      for(Puiss = 128, i = 0 ; i <= 7 ; i++, Puiss /= 2)
         if(Tmp1 & Puiss) sprintf(str + strlen(str), "a%d/", i);
      str[strlen(str) - 1] = 0;
   }
   else if((Allowed == immd) || (Allowed == imm3) || (Allowed == imm4) || (Allowed == imm8))
   {
      int NoNumber = 0;
      tListItem * Item;
      longword Value;
      longword Tmp;
      signed long SValue = 0;
      word PointTo = Code - Buffer->File;
      switch(Allowed)
      {
         case immd: Tmp = ToWord(Code);
                    Code += 2;
                    switch(Size)
                    {
                       case Byte: if(Item = ListImported.Search(Code - Buffer->File - 2))
                                  {
                                     if(!(Item->PointTo & 0x8000))
                                     {
                                        fprintf(stderr, "Erreur : RAM call should be WORD.\n");
                                        exit(1);
                                     }
                                     NoNumber = 1;
                                     strcpy(str, "#");
                                     strcat(str, Item->Label);
                                     ItemToDelete = Item;
                                  }
                                  else
                                  {
                                     Value = Tmp & 0xFF;
                                     SValue = (signed int)((signed char)(Tmp & 0xFF));
                                  }
                                  break;
                       case Word: if(Item = ListImported.Search(Code - Buffer->File - 2))
                                  {
                                     if(!(Item->PointTo & 0x8000))
                                     {
                                        fprintf(stderr, "Erreur : RAM call should be WORD.\n");
                                        exit(1);
                                     }
                                     NoNumber = 1;
                                     strcpy(str, "#");
                                     strcat(str, Item->Label);
                                     ItemToDelete = Item;
                                  }
                                  else
                                  {
                                     Value = Tmp;
                                     SValue = (signed int) (Tmp);
                                  }
                                  break;
                       case Long: if(Item = ListImported.Search(Code - Buffer->File - 2))
                                  {
                                     Code += 2;
                                     strcpy(str, "#");
                                     strcat(str, Item->Label);
                                     NoNumber = 1;
                                     ItemToDelete = Item;
                                  }
                                  else
                                  {
                                     Value = (Tmp << 16) + ToWord(Code);
                                     Code +=2;
                                  }
                                  break;
                    }
                    break;
         case imm3: Value = (Opcode >> 9) & 7; if(!Value) Value = 8; break;
         case imm4: Value = Opcode & 0xF; break;
         case imm8: Value = Opcode & 0xFF; SValue = (signed char)(Opcode & 0xFF); break;
      }
      if(!SValue) SValue = Value;

      if(!NoNumber)
      {
         if(!isrel)
            sprintf(str, "#$%lX", SValue);
         else
         {
            if(!SValue) SValue = Value;
            PointTo += SValue;
            sprintf(str, "%s", GetLabelName(PointTo));
            Referenced = PointTo;
         }
      }
   }

   return Code;
}



// Returns next offset, 0 if end
byte * Disasm(byte * pCode)
{
   char Asm[100];
   int i, Found;
   byte * Code = pCode;
   word OffsetInstruction = Code - Buffer->File, OffsetNext, Index;
   word Opcode = ToWord(Code);
   Code += 2;
   word IsCCBranch = 0;
   Dest = 0; // Indicateur si dest ou pas
   int IsBra = 0, IsBsr = 0;

   Referenced = 0;

   if((Buffer->Flags[OffsetInstruction] & TO_DISAS) || (OffsetInstruction > Buffer->Size) || (ListStopPoints.Search(OffsetInstruction)))
      return NULL;

   Found = 0;
   for(i = 0 ; (i < NbOpcodes) && (!Found) ; i++)
   {
      byte * AncienCode = Code;
      if(InstructionsTable[i].CleanOpcode == (Opcode & InstructionsTable[i].Mask))
      {
         Found = 1;

         ItemToDelete = NULL;

         strcpy(Asm, InstructionsTable[i].Mnemonic);
         Size = InstructionsTable[i].SizeOperandes;
         if(InstructionsTable[i].SizeOperandes & GetSize)
         {
            Size = (Opcode >> 6) & 3;
            switch(Size)
            {
               case 0: strcat(Asm, ".b");
                       Size = Byte;
                       break;
               case 1: strcat(Asm, ".w");
                       Size = Word;
                       break;
               case 2: strcat(Asm, ".l");
                       Size = Long;
                       break;
               case 3: Found = 0;
                       break;
            }
         }

         if(*Asm == 'C')
         {
            char Tmp[10];
            int CC = (Opcode >> 8) & 0xF;
            int i = 2;

            while((Asm[i] != 0) && (Asm[i] != '.')) Tmp[i - 2] = Asm[i++];

            Tmp[i - 2] = 0;
            if((*Tmp == 's') && (CC == 0))
               strcat(Tmp, "t");
            else
               strcat(Tmp, CCodes[CC]);
            strcat(Tmp, &Asm[i]);
            strcpy(Asm, Tmp);

            if(*Asm != 's') IsCCBranch = 1;
         }

         if(*Asm == 'F')
         {
            char Tmp[10];
            char Flow = (Opcode & 256)?'l':'r';
            int i = 1;

            while((Asm[i] != 0) && (Asm[i] != '.')) Tmp[i - 1] = Asm[i++];

            Tmp[i - 1] = Flow;
            while(Asm[i] != 0) Tmp[i] = Asm[i++];
            Tmp[i] = 0;
            strcpy(Asm, Tmp);
         }

         strcat(Asm, "\t");
         if(Found && (InstructionsTable[i].NbOperandes >= 1))
         {
            Dest = 0;
            Code = Decode(Asm + strlen(Asm), Code, Opcode, InstructionsTable[i].AllowedForSource, &Found);
         }
         if(Found && (InstructionsTable[i].NbOperandes == 2))
         {
            Dest = 1;
            strcat(Asm, ",");
            Code = Decode(Asm + strlen(Asm), Code, Opcode, InstructionsTable[i].AllowedForDest, &Found);
         }

      }
      if(!Found) Code = AncienCode;
   }
   i --;

   if(ItemToDelete)
      ListImported.Rm(ItemToDelete);

   if(!Found)
      sprintf(Asm, "dc.w\t$%04X\t; Illegal", ToWord(Code - 2));


   if(strncmp(Asm, "bra.w", 5) == 0)
   {
      char Tmp[100];
      sprintf(Tmp, "dc.w\t$6000\n\tdc.w\t%s-*\t; %s", Asm + 6, Asm);
      strcpy(Asm, Tmp);
      IsBra = 1;
   }
   if(strncmp(Asm, "bsr.w", 5) == 0)
   {
      char Tmp[100];
      sprintf(Tmp, "dc.w\t$6100\n\tdc.w\t%s-*\t; %s", Asm + 6, Asm);
      strcpy(Asm, Tmp);
      IsBsr = 1;
   }

   ListDisas.Add(TypeDisas, OffsetInstruction, Asm, Code - pCode);

   OffsetNext = Code - Buffer->File;
   for(Index = OffsetInstruction; Index < OffsetNext ; Index++)
   {
      Buffer->Flags[Index] |= TO_DISAS;
   }

//   static Counter = 0;
//   Counter++;
//   if(Counter > 200) return NULL;


   if(strncmp(Asm, "rts", 3) == 0) return NULL;
   else if(IsBra)
   {
      ListEntryPoints.Add(TypeEntryPoint, Referenced, GetLabelName(Referenced), 0);
      return NULL;
   }
   else if(IsBsr)
      ListEntryPoints.Add(TypeEntryPoint, Referenced, GetLabelName(Referenced), 0);
   else if(strncmp(Asm, "jmp", 3) == 0)
   {
      if(Referenced) ListEntryPoints.Add(TypeEntryPoint, Referenced, GetLabelName(Referenced), 0);
      return NULL;
   }
   else if(strncmp(Asm, "jsr", 3) == 0)
   {
      if(Referenced) ListEntryPoints.Add(TypeEntryPoint, Referenced, GetLabelName(Referenced), 0);
   }
   else if(strncmp(Asm, "bra", 3) == 0)
   {
      ListEntryPoints.Add(TypeEntryPoint, Referenced, GetLabelName(Referenced), 0);
      return NULL;
   }
   else if(strncmp(Asm, "bsr", 3) == 0)
   {
      ListEntryPoints.Add(TypeEntryPoint, Referenced, GetLabelName(Referenced), 0);
   }
   else if(IsCCBranch)
      ListEntryPoints.Add(TypeEntryPoint, Referenced, GetLabelName(Referenced), 0);



   return Code;
}
