#include "TI92dasm.h"

List::List()
{
   First = NULL;
   Last = NULL;
   NbItems = 0;
}

void List::Add(tItemType pType, word pOffset, char * pLabel, word pPointTo)
{
 char * Label = NULL;

 if(pLabel)
 {
    Label = (char *)Errmalloc(strlen(pLabel) + 1);
    strcpy(Label, pLabel);
 }

 if(!First)
 {
  First = (tListItem *)Errmalloc(sizeof(tListItem));
  Last = First;
 }
 else
 {
  Last->Next = (tListItem *)Errmalloc(sizeof(tListItem));
  Last = Last->Next;
 }

 Last->Next = NULL;
 Last->Type = pType;
 Last->Offset = pOffset;
 Last->Label = Label;
 Last->PointTo = pPointTo;

 NbItems++;
}

tListItem * List::Search(word pOffset)
{
   tListItem * Ret;

   Ret = First;

   while((Ret->Offset != pOffset) && (Ret != NULL)) Ret = Ret->Next;

   return Ret;
}

void List::Print(char * str)
{
   tListItem * Item = First;
   size_t n;

   while(Item != NULL)
   {
      switch(Item->Type)
      {
         case TypeLabel      :  n = strlen(Item->Label);
                                printf("%s%s:\t", str, Item->Label);
                                if(n < 8) printf("\t");
                                if(n < 16) printf("\t");
                                if(n < 24) printf("\t");
                                if(n < 32) printf("\t");
                                printf("; Offset : $%04X\n", Item->Offset);
                                break;
         case TypeBSS        :
         case TypeROM_CALL   :
         case TypeRAM_CALL   :
         case TypeImported   : printf("%sImported item : %s at offset $%04X, points to $%04X\n", str, Item->Label, Item->Offset, Item->PointTo);
                               break;

         case TypeReloc      : printf("%sReloc item at : $%04X that points to : $%04X = label : %s\n", str, Item->Offset, Item->PointTo, Item->Label);
                               break;

         case TypeEntryPoint : printf("%sEntry point at offset : $%04X, label : %s\n", str, Item->Offset, Item->Label);
                               break;
         case TypeDisas      : n = strlen(strchr(Item->Label, '\t') + 1);
                               printf("%s\t%s\t", str, Item->Label);
                               if(n < 8) printf("\t");
                               if(n < 16) printf("\t");
                               printf("; Offset : $%04X : ", Item->Offset);
                               for(int i = 0 ; i < Item->PointTo / 2; i++)
                                  printf("%04X ", ToWord(Buffer->File + Item->Offset + i * 2));
                               printf("\n");
                               break;
         case TypeData       : n = strlen(strchr(Item->Label, '\t') + 1);
                               printf("%s\t%s\t", str, Item->Label);
                               if(n < 8) printf("\t");
                               if(n < 16) printf("\t");
                               printf("; Offset : $%04X\n", Item->Offset);
                               break;
         case TypeTexte      : printf("%s%s\n", str, Item->Label);
                               break;
      }
      Item = Item->Next;
   }
}

tListItem * List::GetNext(tListItem * Current, word pType)
{
   tListItem * Ret;
   if(Current == NULL)
      Ret = First;
   else
      Ret = Current->Next;

   if(pType == TypeAny) return Ret;

   while((Ret != NULL) && (Ret->Type != pType)) Ret = Ret->Next;
   return Ret;
}

void List::SortFollowingToOffset()
{
   int HaveSwaped;
   tListItem * Current, * Previous, * pTmp;
   tListItem Tmp;
   word PreviousOffset;
   fprintf(stderr, "Sorting (number of swaps is displayed for each iteration) :\n");

   do
   {
      HaveSwaped = 0;
      PreviousOffset = 0;
      Previous = NULL;
      Current = First;

      while(Current)
      {
         if(Current->Offset < PreviousOffset)
         {
            // Swap
            Tmp = *Current;
            *Current = *Previous;
            *Previous = Tmp;

            pTmp = Previous->Next;
            Previous->Next = Current;
            Current->Next = pTmp;
            HaveSwaped++;
         }
         PreviousOffset = Current->Offset;
         Previous = Current;
         Current = Current->Next;
      }
      fprintf(stderr, "\r%d   ", HaveSwaped);
   } while(HaveSwaped);
   fprintf(stderr, "\rEnd of sort.\n");
}

void List::Merge(List TheList)
{
   tListItem * Current = NULL;
   while(Current = TheList.GetNext(Current))
   {
      Add(Current->Type, Current->Offset, Current->Label, Current->PointTo);
      TheList.Rm(NULL);
      Current = NULL;
   }

}

int List::card()
{
 return NbItems;
}

int List::Rm(tListItem * WhichOne)
{
   int Ret = 0;

   if(First != NULL)
   {
      if((WhichOne == NULL) || (First == WhichOne))
      {
        tListItem * Tmp = First->Next;

         free(First->Label);
         free(First);
         First = Tmp;
         NbItems--;
         Ret = 1;
      }
      else
      {
         tListItem * Tmp = First->Next;
         tListItem * Previous = First;
         while((Tmp != WhichOne) && Tmp)
         {
            Previous = Tmp;
            Tmp = Tmp->Next;
         }
         if(Tmp != NULL)
         {
            Previous->Next = Tmp->Next;
            free(Tmp->Label);
            free(Tmp);
            NbItems--;
            Ret = 1;
         }
      }
   }

   tListItem * Item = First;
   if(!First)
      Last = NULL;
   else
   {
      while(Item->Next) Item = Item->Next;
      Last = Item;
   }

   return  Ret;
}

void List::Free()
{
 while(Rm(NULL));
}